/*******************
Sentech USB cameras

Author: Marius Silaghi <msilaghi@fit.edu>
January 2014

This program shows how to identify which physical 
device corresponds to which "/dev/videoX" file, 
when multiple cameras of identical type
are connected to the same computer.

The program finds the "sysfs" style name of the
USB port to which each /dev/videoX device is connected.

If the camera with a given function is always connected 
to the same USB port, then it will always receive the 
same sysfs name, and can be identified by applications.

*******************/

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h> 
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <linux/videodev2.h>
#include <Sentech/sentech.h>

extern char *optarg;
unsigned int category = 0x000058, category_inside = 0x005C, data = 0;
int out = 0;
int direct = 0;

int main(int argc, char**argv){
  char* crt_dev, * pattern = "/dev/video";
  int video_fd;
  int opt, verbose=0, flags=0, i=0, retval=0;
  struct sentech_fpga2 cmd;

  while ((opt = getopt(argc, argv, "hd:D:rsc:i:vtl")) != -1) {
    switch (opt) {
    case 's': // out (write)
      out = 1;
      break;
    case 'l': // load offset
      out = 2;
      break;
    case 't': // direct
      direct = 1;
      break;
    case 'r': // read from camera
      out = 0;
      break;
    case 'd': // device
      pattern = optarg;
      break;
    case 'D': // data to write
      data = strtoul(optarg, NULL, 0);
      break;
    case 'i':
      category_inside = strtoul(optarg, NULL, 0);
      break;
    case 'c':
      category = strtoul(optarg, NULL, 0);
      break;
    case 'v':
      verbose = 1;
      break;
    case 'h':
    default: /* '?' */
      fprintf(stderr, "Usage: %s [-d device] [-v] [-c category_offset] [-i inside_category_offset] [-D data] -s -r -l -t\n"
	      "\t-v verbose\n"
	      "\t-s write the data in the camera\n"
	      "\t-r receive the data from the camera\n"
	      "\t-l load offset in category (place the old value in inside_category).\n"
	      "\t     The request is synchronous adding switch '-i 1' and asynchronous with '-i 0'. Default is synchronous.\n"
	      "\t-t direct access to the address in category without using inside category\n"
	      "\t\t   (else previously loaded offset of category is added to inside)\n"
	      "\n\t You can use hexa like 0x00058 for category_offset, inside_category_offset, or Data to write there\n",
	      argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  cmd.category_offset = category;
  cmd.inside_category_offset = category_inside;
  cmd.data = data;
  cmd.direct = direct;
  cmd.filled = 0;
  

  if (verbose)
    printf("Found device %s\n", pattern);
  video_fd = open(pattern, O_RDONLY|O_NONBLOCK);

  if (video_fd == -1) {
    if (! verbose)
      fprintf(stderr, "Found device %s\n", pattern);
    perror ("Problem opening video device:");
    return;
  }
 
  if (out == 2) {
    if (ioctl(video_fd, SENTECH_IOC_LOAD_OFFSET, &cmd) == -1) {
      perror("Query LOAD_CATEGORY fails:");
      if (! verbose)
        fprintf(stderr, "Using device %s offset %x\n", pattern, cmd.inside_category_offset);
    } else {
     printf("Device %s request got: data=%x fill=%d stat=%d offset=%08xh\n", pattern, cmd.data, cmd.filled, cmd.status, cmd.inside_category_offset);
    }
  } 
  if (out == 1) {
    if (ioctl(video_fd, SENTECH_IOC_S_CATEGORY, &cmd) == -1) {
      perror("Query S_CATEGORY fails:");
      if (! verbose)
	fprintf(stderr, "Using device %s\n", pattern);
    } else {
      printf("Device %s request sent\n", pattern);
    }
  } 
  if (out == 0) {
    if (ioctl(video_fd, SENTECH_IOC_G_CATEGORY, &cmd) == -1) {
       perror("Query G_CATEGORY fails:");
       if (! verbose)
	 fprintf(stderr, "Using device %s\n", pattern);
    } else {
      printf("Device %s request got: data=%x fill=%d stat=%d\n", pattern, cmd.data, cmd.filled, cmd.status);
    }
  }

  close(video_fd);
}
