/*******************
Sentech USB cameras

Author: Marius Silaghi <msilaghi@fit.edu>
January 2011

This program shows how to identify which physical 
device corresponds to which "/dev/videoX" file, 
when multiple cameras of identical type
are connected to the same computer.

The program finds the "sysfs" style name of the
USB port to which each /dev/videoX device is connected.

If the camera with a given function is always connected 
to the same USB port, then it will always receive the 
same sysfs name, and can be identified by applications.

*******************/

#include <glob.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h> 
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/videodev2.h>
#include <Sentech/sentech.h>

#ifndef STC_PRODUCT_PORT
#define STC_PRODUCT_PORT 1
#endif

extern char *optarg;

int errfunc (const char *epath, int eerrno) {
  fprintf(stderr,"Error %s (%d) accessing $s.", 
	  strerror(eerrno), eerrno, epath);
  return 0;
}

int main(int argc, char**argv){
  char* crt_dev, * pattern = "/dev/video*";
  glob_t globbuf;
  int opt, verbose=0, flags=0, i=0, retval=0;
  struct v4l2_querymenu arg;
  arg.id = V4L2_CID_STC_PRODUCT;
  arg.index = STC_PRODUCT_PORT;
  while ((opt = getopt(argc, argv, "p:i:x:v")) != -1) {
    switch (opt) {
    case 'p':
      pattern = optarg;
      break;
    case 'i':
      arg.id = atoi(optarg);
      break;
    case 'x':
      arg.index = atoi(optarg);
      break;
    case 'v':
      verbose = 1;
      break;
    default: /* '?' */
      fprintf(stderr, "Usage: %s [-p pattern] [-v] [-i ctrl_id] [-x ctrl_idx]\n",
	      argv[0]);
      exit(EXIT_FAILURE);
    }
  }
  
  globbuf.gl_offs = 0;
  retval = glob(pattern, flags, errfunc, &globbuf);

  switch(retval){
  case GLOB_NOSPACE:
    fprintf(stderr, "Out of memory: too many cameras!\n");
    break;
  case GLOB_ABORTED:
    fprintf(stderr, "Aborted due to read error!\n");
    break;
  case GLOB_NOMATCH:
    fprintf(stderr, "No video device found!\n");
    break;
  default:break;
  }

  for (i = 0; i < globbuf.gl_pathc; i++) {
    int video_fd;
    crt_dev = globbuf.gl_pathv[i];
    if(verbose)
      printf("Found device %d/%d: %s\n", i+1, globbuf.gl_pathc, crt_dev);
    video_fd = open(crt_dev, O_RDONLY|O_NONBLOCK);

    if(video_fd == -1) {
      if(!verbose)
	fprintf(stderr, "Found device %d/%d: %s\n", i+1, globbuf.gl_pathc, crt_dev);
      perror ("Problem opening video device:");
      continue;
    }

    if(ioctl(video_fd, VIDIOC_QUERYMENU, &arg)==-1) {
      if(!verbose)
	fprintf(stderr, "Found device %d/%d: %s\n", i+1, globbuf.gl_pathc, crt_dev);
      perror("Query STC_PRODUCT_PORT fails:");
    } else {
      printf("Device %s is at port \"%s\"\n", crt_dev, arg.name);
    }

    close(video_fd);
  }
  globfree(&globbuf);
}
