/*
Pulse output on a Sentech USB camera.
Marius Silaghi <msilaghi@fit.edu>

Sample program, showing how to output a pulse on the IO2 output
prior to taking N frames at maximum frame rate. There are 2 
prodecures available here for generating the pulse.

- pulseCT() method generates pulses that are larger than 1ms (due to the use 
of system calls).

-purseTE() generates the requested 5us pulse with a reliable width. However, 
the IO2 mode has to be changed between the pulse generating capture and 
remaining frames. Due to this, the delay between this pulse and the next 
capture may be higher to 1ms due to the use of system calls, but that may be 
less than the time to capture a full frame. The camera needs time to settle 
after the mode-change. Hurrying up to switch the mode after the pulse, 
sometimes loses the mode change message, so that has to be detected and the 
request message has to be regenerated.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <getopt.h>             /* getopt_long() */

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <asm/types.h>          /* for videodev2.h */

#include <linux/videodev2.h>

#include <Sentech/sentech.h>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

#define G_SYSERROR2(a) printf("syserr2: " a "\n")
#define G_ERROR2(a) printf("err2: " a "\n")
#define G_LOG2(a) printf("log2: " a "\n")

/* Default VGA parameters for bitmap header */
const char * device = "/dev/video0";
long int height=480;
long int width=640;
// TC83: -h 0x400 -w 0x300
// TC152: -h 1024 -w 1360

void mydelay(int us){
  struct timeval tv;
  tv.tv_sec=0;
  tv.tv_usec=1000;
  select (1, NULL, NULL, NULL, &tv); 
}

static void
errno_exit                      (const char *           s)
{
        fprintf (stderr, "%s error %d, %s\n",
                 s, errno, strerror (errno));

        exit (EXIT_FAILURE);
}

struct buffer {
        void *                  start;
        size_t                  length;
} * buffers;

int real_bufcnt = 4;
struct v4l2_format fmt;
//struct v4l2_control ctrl;
struct v4l2_control arg_ctl;
int result;
void autoLC(int m_DeviceFd, int time){
  struct v4l2_control ctrl;
  struct timeval tv;
  int r;
  ctrl.id = V4L2_CID_STC_TRIGGER_MODE;
  ctrl.value = STC_TRIGGER_MODE_NORMAL;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &ctrl);
  if(result != 0){ perror("ioctl trigger");exit(2);}

  ctrl.id=V4L2_CID_AUTOGAIN;
  ctrl.value=STC_AUTOGAIN_ON;
  ioctl(m_DeviceFd,VIDIOC_S_CTRL,&ctrl);

  sleep(time);

  ctrl.id = V4L2_CID_STC_TRIGGER_MODE;
  ctrl.value = STC_TRIGGER_MODE_TRIGGER;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &ctrl);
  if(result != 0){perror("ioctl trigger");exit(2);}

  ctrl.id=V4L2_CID_AUTOGAIN;
  ctrl.value=STC_AUTOGAIN_OFF;
  ioctl(m_DeviceFd,VIDIOC_S_CTRL,&ctrl);
  tv.tv_sec=0;
  tv.tv_usec=200000;
  r = select (1, NULL, NULL, NULL, &tv); 
  ctrl.id=V4L2_CID_STC_CLEAR_STATE;
  ctrl.value=0;
  ioctl(m_DeviceFd,VIDIOC_S_CTRL,&ctrl); 
}

int pulseTE(int m_DeviceFd, int delay){
  struct timeval tv;
  int cnt=0;
  int retval=0;
 repulse:
  mydelay(100000);
  /*SOFTWARE TRIGGER*/
  arg_ctl.id = V4L2_CID_STC_TRIGGER_SOURCE;
  arg_ctl.value = STC_TRIGGER_SOURCE_SOFTWARE;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl trigger source");exit(2);}
  /*SET TRIGGER_MODE*/
  arg_ctl.id = V4L2_CID_STC_TRIGGER_MODE;
  arg_ctl.value = STC_TRIGGER_MODE_TRIGGER;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl trigger mode");exit(2);}
  /*SET PULSE WIDTH*/
  arg_ctl.id = V4L2_CID_STC_W_TRG_OUT;
  arg_ctl.value = delay*25;//164; //clocks*40ns
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}
  /*SET PULSE TO TRANSFER-END*/
  arg_ctl.id = V4L2_CID_STC_IO2_FUNCTION;
  arg_ctl.value = IO_O_TRAN_END;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}

  mydelay(100000);

  /*DO TRIGGER*/
  arg_ctl.id = V4L2_CID_STC_TRIGGER;
  arg_ctl.value = 1;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}
  
  fd_set fds;
  FD_ZERO (&fds);
  FD_SET (m_DeviceFd, &fds);
  tv.tv_sec=0;
  tv.tv_usec=100000;
  int r = select (m_DeviceFd+1, &fds, NULL, NULL, &tv);
	
  if ( r <= 0 ) {
    fprintf (stderr,"(-P,%d)",cnt);
    retval=-1;
  }else{
    retval=0;
    int buftype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    struct v4l2_buffer buf;
    int res;
    res=ioctl(m_DeviceFd, VIDIOC_DQBUF, &buf);
    if(res!=0){
      fprintf(stderr,"Pulse=%d",errno);
      retval=-1;
    }
    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    ioctl(m_DeviceFd, VIDIOC_QBUF, &buf);
  }

  //mydelay(1000);
  //SOFTWARE TRIGGER
  arg_ctl.id = V4L2_CID_STC_TRIGGER_SOURCE;
  arg_ctl.value = STC_TRIGGER_SOURCE_IGNORE;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl trigger source");exit(2);}
  //mydelay(1000);
  //SET TRIGGER_MODE
  arg_ctl.id = V4L2_CID_STC_TRIGGER_MODE;
  arg_ctl.value = STC_TRIGGER_MODE_NORMAL;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl normal mode");exit(2);}
  //mydelay(1000);
  //SET LEVEL
  arg_ctl.id = V4L2_CID_STC_IO2_FUNCTION;
  arg_ctl.value = IO_O_CT;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  arg_ctl.id = V4L2_CID_STC_IO_STATUS_2;
  arg_ctl.value = 0;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 HIGH function");exit(2);}
  if(retval<0){
    fprintf(stderr,"-");
    if(++cnt<2)
      goto repulse;
  }else
    fprintf(stderr,"+");
}

void pulseCT(int m_DeviceFd, int delay){ 
  struct timeval tv;

  arg_ctl.id = V4L2_CID_STC_IO2_FUNCTION;
  arg_ctl.value = IO_O_CT;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}

  arg_ctl.id = V4L2_CID_STC_IO_STATUS_2;
  arg_ctl.value = 1;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 HIGH function");exit(2);}

  mydelay(delay);
  /*
  tv.tv_sec=0;
  tv.tv_usec=delay;
  int r = select (1, NULL, NULL, NULL, &tv); 
  */

  arg_ctl.id = V4L2_CID_STC_IO_STATUS_2;
  arg_ctl.value = 0;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 LOW function");exit(2);}
}

int main(int argc, char**argv) {
  int opt, saving=0, images=50, hard_trigger=0, auto_wb_lc=0, bpp=8;
  while ((opt = getopt(argc, argv, "d:n:h:w:a:RHs")) != -1) {
    switch(opt){
    case 'd': device = optarg; break;
    case 'h': height = atoi(optarg); break;
    case 'w': width = atoi(optarg); break;
    case 'n': images = atoi(optarg); break;
    case 'H': hard_trigger = 1; break;
    case 'a': auto_wb_lc = 1; break;
    case 'R': bpp = 8; break;
    case 's': saving = 1; break;
    default: fprintf(stderr, "%s -d device -h height -w width -n 10\n", argv[0]); exit(-1);
    }
  }

  int m_DeviceFd = open(device, O_RDWR, O_NONBLOCK);
  if(m_DeviceFd<0) errno_exit("open");

  /* Set White Balance, Auto Luminance choice */
  /*
  arg_ctl.id=V4L2_CID_AUTO_WHITE_BALANCE;
  arg_ctl.value=STC_AUTO_WB_OFF;
  ioctl(m_DeviceFd,VIDIOC_S_CTRL,&arg_ctl);
  arg_ctl.id=V4L2_CID_AUTOGAIN;
  arg_ctl.value=STC_AUTOGAIN_OFF;
  ioctl(m_DeviceFd,VIDIOC_S_CTRL,&arg_ctl);
  */
  fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width       = width; 
  fmt.fmt.pix.height      = height;
  if(bpp>8) fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
  else fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SRGGB8;
  //fmt.fmt.pix.priv        = 1;
  ioctl(m_DeviceFd, VIDIOC_S_FMT, &fmt);

  /*OUTPUT*/
  arg_ctl.id = V4L2_CID_STC_IO2_FUNCTION;
  arg_ctl.value = hard_trigger?IO_I_TRIG:IO_O_CT;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}
  /*LOW OUTPUT*/
  arg_ctl.id = V4L2_CID_STC_IO_STATUS_2;
  arg_ctl.value = 0;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 HIGH function");exit(2);}

  /*
  arg_ctl.id = V4L2_CID_STC_TRIGGER_SOURCE;
  arg_ctl.value = hard_trigger?
    STC_TRIGGER_SOURCE_HARDWARE:STC_TRIGGER_SOURCE_SOFTWARE;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl trigger source");exit(2);}
  //
  arg_ctl.id = V4L2_CID_STC_IO2_FUNCTION;
  //arg_ctl.value = hard_trigger?IO_I_TRIG:IO_O_CT;
  arg_ctl.value = hard_trigger?IO_I_TRIG:IO_O_TRAN_END;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl IO2 function");exit(2);}
  //for IO_O_CT
  arg_ctl.id = V4L2_CID_STC_TRIGGER_MODE;
  arg_ctl.value = STC_TRIGGER_MODE_TRIGGER;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl trigger mode");exit(2);}
  //
  arg_ctl.id = V4L2_CID_STC_CNT_RST;
  arg_ctl.value = 0;
  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
  if(result != 0){perror("ioctl counter reset");exit(2);}
  */
  int ioctl_result;
  struct v4l2_requestbuffers req;
  CLEAR (req);
  req.count               = real_bufcnt;
  req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory              = V4L2_MEMORY_MMAP;
  ioctl_result = ioctl (m_DeviceFd, VIDIOC_REQBUFS, &req);
  printf("REQBUF result %d, bufs=%d\n",ioctl_result,req.count);
  real_bufcnt = req.count;

  buffers = (struct buffer*)calloc (req.count, sizeof (*buffers));
  for (int n_buffers = 0; n_buffers < req.count; ++n_buffers) {
    struct v4l2_buffer buf;
    CLEAR (buf);
    buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory      = V4L2_MEMORY_MMAP;
    buf.index       = n_buffers;
    if (-1 == ioctl (m_DeviceFd, VIDIOC_QUERYBUF, &buf))
      errno_exit ("VIDIOC_QUERYBUF");
    buffers[n_buffers].length = buf.length;
    buffers[n_buffers].start =
      mmap (NULL /* start anywhere */,
	    buf.length,
	    PROT_READ | PROT_WRITE /* required */,
	    MAP_SHARED /* recommended */,
	    m_DeviceFd, buf.m.offset);
    if (MAP_FAILED == buffers[n_buffers].start)
      errno_exit ("mmap");
  }

  int buftype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  struct v4l2_buffer buf;
  int res;
  
  CLEAR(buf);
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  for (unsigned int i = 0; i < real_bufcnt; i++){
    buf.index = i;
    if (ioctl(m_DeviceFd, VIDIOC_QBUF, &buf) == -1){
      printf("Error: %s buf(%d/%d)\n", strerror(errno),i,real_bufcnt);
      G_ERROR2("Image grab failed - queue buf");
      break;
    }else{
      printf("Success:in QBUF %d/%d\n",i,real_bufcnt);
    }
  }
  if (ioctl(m_DeviceFd, VIDIOC_STREAMON, &buftype) == -1) {
    G_ERROR2("Image grab failed - streamon");
    exit(-1);
  }    

  unsigned int autoLCc=0;
  for(;;){
  pulses:
    if(auto_wb_lc) {
      //if(autoLCc%5<2) 
	autoLC(m_DeviceFd,1);
// else sleep(1); 
      autoLCc ++;
    }
    pulseCT(m_DeviceFd, 5);
    G_LOG2("PULSE!!!");
    for(int n=0; n<images; n++){
      /* number of attempts in case wait is interrupted */
      for ( int count=0; count<10; count++ ) {
	fd_set fds;
	struct timeval tv;
	int r;
	
	FD_ZERO (&fds);
	FD_SET (m_DeviceFd, &fds);
	
	// set timeouts
	tv.tv_sec=0;
	tv.tv_usec=500000;
	
	r = select (m_DeviceFd+1, &fds, NULL, NULL, &tv);
	
	if ( -1 == r ) {
	  if ( EINTR == errno ) {
	    G_LOG2 ("Capture Interrupted. Retry...\n");
	    continue;
	  }
	}
	
	if ( 0 == r ) {
	  //G_SYSERROR2 ( "SELECT Timeout on device, restarting camera module" );
	  fprintf(stderr,"-");
	  /*SOFTWARE TRIGGER*/
	  arg_ctl.id = V4L2_CID_STC_TRIGGER_SOURCE;
	  arg_ctl.value = STC_TRIGGER_SOURCE_IGNORE;
	  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
	  if(result != 0){perror("ioctl trigger source");exit(2);}
	  /*SET TRIGGER_MODE*/
	  arg_ctl.id = V4L2_CID_STC_TRIGGER_MODE;
	  arg_ctl.value = STC_TRIGGER_MODE_NORMAL;
	  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
	  if(result != 0){perror("ioctl normal mode");exit(2);}
	  /*SET REFRESH*/
	  arg_ctl.id = V4L2_CID_STC_REFRESH;
	  arg_ctl.value = 0;
	  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
	  if(result != 0){perror("ioctl normal mode");exit(2);}
	  /*
	  arg_ctl.id = V4L2_CID_STC_TRIGGER;
	  arg_ctl.value = 0;
	  result = ioctl(m_DeviceFd, VIDIOC_S_CTRL, &arg_ctl);
	  if(result != 0){perror("ioctl fps");exit(2);}
	  */
	  goto pulses;
	  continue;
	}
	// With O_NONBLOCK flag on open(), VIDIOC_DQBUF returns immediately
	// with an EAGAIN error code when no buffer is available
	while ((res = ioctl(m_DeviceFd, VIDIOC_DQBUF, &buf)) < 0 &&
	       ((errno == EAGAIN) || (errno == EINTR)));
	if (res < 0) {
	  G_ERROR2("Image grab failed - dequeue");
	  ioctl(m_DeviceFd, VIDIOC_QBUF, &buf);
	exit(-1); // may not be required
	} else {
	  static int cnt=0;
	  char name[]="last_0000000000_00000000.bmp\0";
	  sprintf(name,"snap_%010i_%08i.rbbg",cnt++,buf.sequence);
	  /*
	  printf("Got seq=%i idx=%i used=%i flags=%i time=%ld.%.6ld\n",
		 buf.sequence,
		 buf.index, buf.bytesused,
		 buf.flags, buf.timestamp.tv_sec, buf.timestamp.tv_usec);
	  */
	  fprintf(stderr,".");
	  if(saving){
	    
	    int storage=open(name,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	    if(write(storage, buffers[buf.index].start, buffers[buf.index].length)!=buffers[buf.index].length)
	      G_LOG2("BITMAP data writing error");
	    close(storage);
	  }
	  ioctl(m_DeviceFd, VIDIOC_QBUF, &buf);
	}
	break;
      }
    }
  }
  if (ioctl(m_DeviceFd, VIDIOC_STREAMOFF, &buftype) == -1) {
    G_ERROR2("Image grab failed - streamoff");
    //exit(-1);
  }
}
